from __future__ import unicode_literals
from mogwai._compat import integer_types, float_types, string_types, iteritems
from mogwai.tools import LazyImportClass
from mogwai.exceptions import MogwaiException

element = LazyImportClass('mogwai.models.element.Element')  # avoid circular import
numeric_types = (integer_types + float_types)


class Row(object):
    """ Row

    A row represent a table row, from which it's columns can be accessed like a tuple or dict.
    Rows are read-only and accept elements or dicts with as initializers as a result of a GremlinTable query.
    Also the . getattr notation can be used to access elements

    Example:
    row = Row({'person': Friend.create(....), 'myval': 3})
    print "{}:{} - {}".format(row.friend_edge.nickname, row.person.name, row.myval)
    """

    __ready = False
    __okay_setattr = '_Row__'

    def __init__(self, data):
        if isinstance(data, element.klass):  # this class is cached, so no performance hit here.
            data = data.as_dict()
        elif not isinstance(data, dict):
            raise MogwaiException("Result data is not tabular!")
        self.__data = data
        for k, v in data.items():
            if isinstance(k, string_types):
                setattr(self, k, v)
        self.__position = 0
        self.__ready = True

    def __getslice__(self, i, j):
        return list(self.__data.values())[i:j]

    def __setslice__(self, i, j, sequence):
        raise MogwaiException("Row is not editable")

    def __delslice__(self, i, j):
        raise MogwaiException("Row is not editable")

    def __getitem__(self, item):
        if isinstance(item, numeric_types) or isinstance(item, slice):
            return list(self.__data.values())[item]
        return self.__data[item]

    def __setitem__(self, key, value):
        raise MogwaiException("Row is not editable")

    def __delitem__(self, key):
        raise MogwaiException("Row is not editable")

    def __setattr__(self, key, value):
        if not self.__ready or key.startswith(self.__okay_setattr):  # only allow 'private' fields to be set
            super(Row, self).__setattr__(key, value)
        else:
            raise MogwaiException("Row is not editable")

    def __delattr__(self, item):
        raise MogwaiException("Row is not editable")

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def keys(self):
        return self.__data.keys()

    def values(self):
        return self.__data.values()

    def items(self):
        return self.__data.items()

    def iteritems(self):
        for k, v in self.__data.items():
            yield k, v

    def next(self):
        if self.__position == len(self.__data):
            self.__position = 0
            raise StopIteration()
        tmp = list(self.__data.values())[self.__position]
        self.__position += 1
        return tmp

    def __len__(self):
        return len(self.__data)

    def __repr__(self):
        result = "{}(".format(self.__class__.__name__)
        for k, v in iteritems(self.__data):
            result += "{}={}, ".format(k, v)
        result = result.rstrip(", ")
        result += ")"
        return result

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        else:
            return self.__data == other._Row__data


class Table(object):
    """ Table

    A table accepts the results of a GremlinTable in it's constructor.
    It can be iterated over like a normal list, but within the rows
    the dictionaries are accessible via .notation

    For example:

    # returns a table of people & my friend edge to them
    # the edge contains my nickname for that person
    friends = mogwai.gremlin.GremlinTable()

    def get_friends_and_my_nickname(self):
        result = self.friends()
        for i in result:
            print "{}:{}".format(i.friend_edge.nickname, i.person.name)
    """

    def __init__(self, gremlin_result):
        if gremlin_result == [[]]:
            gremlin_result = []

        self.__gremlin_result = gremlin_result
        self.__position = 0

    def __getitem__(self, item):
        """
        Returns an enhanced dictionary
        """
        return Row(self.__gremlin_result[item])

    def __setitem__(self, key, value):
        raise MogwaiException("Cannot edit Table result")

    def __delitem__(self, key):
        raise MogwaiException("Cannot edit Table result")

    def __getslice__(self, i, j):
        return [Row(r) for r in self.__gremlin_result[i:j]]

    def __setslice__(self, i, j, sequence):
        raise MogwaiException("Cannot edit Table result")

    def __delslice__(self, i, j):
        raise MogwaiException("Cannot edit Table result")

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if self.__position == len(self.__gremlin_result):
            self.__position = 0
            raise StopIteration()
        tmp = self.__gremlin_result[self.__position]
        self.__position += 1
        return Row(tmp)

    def __len__(self):
        return len(self.__gremlin_result)

    def __repr__(self):
        return '{}(rows={})'.format(self.__class__.__name__, len(self.__gremlin_result))


__all__ = ['Table', 'Row']


from __future__ import unicode_literals
from nose.plugins.attrib import attr
from copy import deepcopy
from mogwai._compat import print_
from mogwai.models import Vertex, Edge
from mogwai.properties import String, Integer
from mogwai.exceptions import MogwaiRelationshipException
from mogwai.tests.base import BaseMogwaiTestCase, TestVertexModel, counter
from mogwai.relationships.base import Relationship


class TestRelationshipEdgeModel(Edge):
    label = 'test_relationship_edge_model'

    name = String(default='test_edge')
    test_val = Integer(default=counter)


class TestRelationshipVertexModel(Vertex):
    element_type = 'test_relationship_vertex_model'

    name = String(default='test_vertex')
    test_val = Integer(default=counter)

    relation = Relationship(TestRelationshipEdgeModel, TestVertexModel)


@attr('unit', 'relationship')
class GraphRelationshipVertexIOTestCase(BaseMogwaiTestCase):
    """ Test Relationship Vertex IO Functionality """

    @classmethod
    def setUpClass(cls):
        super(GraphRelationshipVertexIOTestCase, cls).setUpClass()
        cls.relationship_base_cls = Relationship
        cls.edge_model = TestRelationshipEdgeModel
        cls.vertex_model = TestRelationshipVertexModel

    def test_instantiation(self):
        """ Test that the Relationship is properly Instantiated """

        v1 = self.vertex_model.create(name='test relationship')
        # setup relationship
        self.assertIsNotNone(v1.relation.top_level_vertex_class)
        self.assertIsNotNone(v1.relation.top_level_vertex)
        self.assertEqual(v1.relation.top_level_vertex, v1)
        v1.delete()

    def test_follow_through(self):
        """ Test that the Relationship property functions """

        v1 = self.vertex_model.create(name='test relationship')
        e1, v2 = v1.relation.create(vertex_params={'name': 'new relation'})

        e1q = v1.outE(TestRelationshipEdgeModel)[0]
        v2q = v1.outV(TestRelationshipEdgeModel)[0]

        self.assertEqual(e1, e1q)
        self.assertEqual(v2, v2q)

        e1.delete()
        v1.delete()
        v2.delete()

    @attr('relationship_isolation')
    def test_relationship_isolation(self):
        """ Test that the relationship adheres to instance methods """

        v11 = self.vertex_model.create(name='test1')
        e1, v12 = v11.relation.create(vertex_params={'name': 'new_relation_1'})

        r11 = deepcopy(v11.relation.vertices())
        print_("Vertex 1-1 relationships: {}".format(r11))

        v21 = self.vertex_model.create(name='test2')
        e2, v22 = v21.relation.create(vertex_params={'name': 'new_relation_2'})

        r2 = deepcopy(v21.relation.vertices())
        print_("Vertex 2-1 relationships: {}".format(r2))

        with self.assertRaises(AssertionError):
            self.assertListEqual(r11, r2)

        r12 = deepcopy(v11.relation.vertices())
        print_("Vertex 1-1 relationships again: {}".format(r12))
        with self.assertRaises(AssertionError):
            self.assertListEqual(r2, r12)

        self.assertListEqual(r11, r12)

        v11.delete()
        v12.delete()
        v21.delete()
        v22.delete()


import copy
import datetime
from calendar import timegm
from decimal import Decimal as _D
import re
import time
import warnings
from mogwai._compat import text_type, string_types, float_types, integer_types, long_, PY3

from uuid import uuid1, uuid4
from uuid import UUID as _UUID

from .base import GraphProperty
from .validators import *


class String(GraphProperty):
    """
    String/CharField property
    """
    data_type = "String"
    validator = string_validator

    def __init__(self, *args, **kwargs):
        required = kwargs.get('required', False)
        self.min_length = kwargs.pop('min_length', 1 if required else None)
        self.max_length = kwargs.pop('max_length', None)
        self.encoding = kwargs.pop('encoding', 'utf-8')
        if 'default' in kwargs and isinstance(kwargs['default'], string_types):
            if not PY3:
                kwargs['default'] = kwargs['default'].encode(self.encoding)
        super(String, self).__init__(*args, **kwargs)

    def to_python(self, value):
        if value is not None:
            if isinstance(value, (bytes, bytearray)) and not isinstance(value, str):
                return value.decode(self.encoding)
            else:
                return value

    def validate(self, value):
        # Make sure it gets encoded properly
        if isinstance(value, text_type) and not PY3:
            value = value.encode(self.encoding)

        value = super(String, self).validate(value)

        # this should never happen unless the Validator is changed
        if value is None:
            return None

        if self.max_length:
            if len(value) > self.max_length:
                raise ValidationError('{} is longer than {} characters'.format(self.property_name, self.max_length))
        if self.min_length:
            if len(value) < self.min_length:
                raise ValidationError('{} is shorter than {} characters'.format(self.property_name, self.min_length))
        return value

Text = String


class Short(GraphProperty):
    """
    Short Data property type
    """
    data_type = "Short"
    validator = integer_validator

    def to_python(self, value):
        if value is not None:
            return int(value)

    def to_database(self, value):
        value = super(Short, self).to_database(value)
        if value is not None:
            return int(value)


class Integer(GraphProperty):
    """
    Integer Data property type
    """
    data_type = "Integer"
    validator = long_validator

    def to_python(self, value):
        if value is not None:
            try:
                return int(value)
            except:
                return long_(value)

    def to_database(self, value):
        value = super(Integer, self).to_database(value)
        if value is not None:
            return long_(value)


class PositiveInteger(Integer):
    """
    Positive Integer Data property type
    """
    data_type = "Integer"
    validator = positive_integer_validator

    def to_python(self, value):
        if value is not None:
            return long_(value)

    def to_database(self, value):
        value = super(Integer, self).to_database(value)
        if value is not None:
            return long_(value)


class Long(GraphProperty):
    """
    Long Data property type
    """
    data_type = "Long"
    validator = long_validator

    def to_python(self, value):
        if value is not None:
            return long_(value)

    def to_database(self, value):
        value = super(Long, self).to_database(value)
        if value is not None:
            return long_(value)


class PositiveLong(Long):
    """
    Positive Long Data property type
    """
    data_type = "Long"
    validator = positive_integer_validator

    def to_python(self, value):
        if value is not None:
            return long_(value)

    def to_database(self, value):
        value = super(Long, self).to_database(value)
        if value is not None:
            return long_(value)


class DateTimeNaive(GraphProperty):
    """
    DateTime Data property type
    """
    data_type = "Double"
    validator = datetime_validator

    def __init__(self, strict=True, **kwargs):
        """
        Initialize date-time column with the given settings.

        :param strict: Whether or not to attempt to automatically coerce types
        :type strict: boolean

        """
        self.strict = strict
        super(DateTimeNaive, self).__init__(**kwargs)

    def to_python(self, value):
        if isinstance(value, datetime.datetime):
            return value
        return datetime.datetime.fromtimestamp(float(value))

    def to_database(self, value):
        value = super(DateTimeNaive, self).to_database(value)
        if value is None:
            return
        if not isinstance(value, datetime.datetime):
            if not self.strict and isinstance(value, string_types + integer_types + float_types):
                value = datetime.datetime.fromtimestamp(float(value))
            else:
                raise ValidationError("'{}' is not a datetime object".format(value))

        tmp = time.mktime(value.timetuple())  # gives us a float with .0
        # microtime is a 6 digit int, so we bring it down to .xxx and add it to the float TS
        tmp += float(value.microsecond) / 1000000.0
        return tmp


class DateTime(GraphProperty):
    """
    UTC DateTime Data property type
    """
    data_type = "Double"
    validator = datetime_utc_validator

    def __init__(self, strict=True, **kwargs):
        """
        Initialize date-time column with the given settings.

        :param strict: Whether or not to attempt to automatically coerce types
        :type strict: boolean

        """
        self.strict = strict
        super(DateTime, self).__init__(**kwargs)

    def to_python(self, value):
        try:
            if isinstance(value, datetime.datetime):
                if value.tzinfo == utc:
                    return self.validator(value)  # .astimezone(tz=utc)
                else:
                    return self.validator(value).astimezone(tz=utc)
        except:  # pragma: no cover
            # this shouldn't happen unless the validator has changed
            pass
        return datetime.datetime.utcfromtimestamp(float(value)).replace(tzinfo=utc)

    def to_database(self, value):
        value = super(DateTime, self).to_database(value)
        if value is None:
            return
        if not isinstance(value, datetime.datetime):
            if isinstance(value, string_types + integer_types + float_types):
                value = datetime.datetime.utcfromtimestamp(float(value)).replace(tzinfo=utc)
            else:
                raise ValidationError("'{}' is not a datetime object".format(value))

        tmp = timegm(value.utctimetuple())  # gives us an integer of epoch seconds without microseconds
        # microtime is a 6 digit int, so we bring it down to .xxx and add it to the float TS
        tmp += float(value.microsecond) / 1000000.0
        return tmp


class UUID(GraphProperty):
    """Universally Unique Identifier (UUID) type - UUID4 by default"""
    data_type = "String"
    validator = validate_uuid4

    def __init__(self, default=lambda: str(uuid4()), **kwargs):
        self.uuid_version = kwargs.pop('version', 'uuid4').lower()
        if self.uuid_version == 'uuid4':
            self.validator = validate_uuid4
        elif self.uuid_version == 'uuid1':
            self.validator = validate_uuid1

        super(UUID, self).__init__(default=default, **kwargs)

    def to_python(self, value):
        val = super(UUID, self).to_python(value)
        if value is not None:
            if isinstance(value, (bytes, bytearray)) and not isinstance(value, str):
                return value.decode('utf-8')
            else:
                return value

    def to_database(self, value):
        val = super(UUID, self).to_database(value)
        if val is None:
            return
        return str(val)


class Boolean(GraphProperty):
    """
    Boolean Data property type
    """
    data_type = "Boolean"
    validator = bool_validator

    def to_python(self, value):
        return bool(value)

    def to_database(self, value):
        val = super(Boolean, self).to_database(value)
        return bool(val)


class Double(GraphProperty):
    """
    Double Data property type
    """
    data_type = "Double"
    validator = float_validator

    def __init__(self, **kwargs):
        self.db_type = 'double'
        super(Double, self).__init__(**kwargs)

    def to_python(self, value):
        if value is not None:
            return float(value)

    def to_database(self, value):
        value = super(Double, self).to_database(value)
        if value is not None:
            return float(value)


class Float(Double):
    """Float class for backwards compatability / if you really want to"""

    def __init__(self, **kwargs):
        warnings.warn("Float type is deprecated. Please use Double.",
                      category=DeprecationWarning)
        super(Float, self).__init__(**kwargs)


class Decimal(GraphProperty):
    """
    Decimal Data property type
    """
    validator = decimal_validator

    def to_python(self, value):
        val = super(Decimal, self).to_python(value)
        if val is not None:
            return _D(val)

    def to_database(self, value):
        val = super(Decimal, self).to_database(value)
        if val is not None:
            return str(val)


class Dictionary(GraphProperty):
    """
    Dictionary Data property type
    """
    data_type = "HashMap"
    validator = dict_validator


class List(GraphProperty):
    """
    List Data property type
    """
    data_type = "ArrayList"
    validator = list_validator


class URL(GraphProperty):
    """
    URL Data property type
    """
    data_type = "String"
    validator = validate_url

    def __init__(self, *args, **kwargs):
        required = kwargs.get('required', False)
        self.min_length = kwargs.pop('min_length', 1 if required else None)
        self.max_length = kwargs.pop('max_length', None)
        self.encoding = kwargs.pop('encoding', 'utf-8')
        if 'default' in kwargs and isinstance(kwargs['default'], string_types):
            if not PY3:
                kwargs['default'] = kwargs['default'].encode(self.encoding)
        super(URL, self).__init__(*args, **kwargs)

    def validate(self, value):
        # Make sure it gets encoded correctly
        if isinstance(value, text_type) and not PY3:
            value = value.encode(self.encoding)

        value = super(URL, self).validate(value)

        # this should never happen unless the validator is customized
        if value in (None, [], (), {}):  # pragma: no cover
            return None

        if value is not None:
            if isinstance(value, (bytes, bytearray)) and not isinstance(value, str):
                return value.decode(self.encoding)

        if self.max_length:
            if len(value) > self.max_length:
                raise ValidationError('{} is longer than {} characters'.format(self.property_name, self.max_length))
        if self.min_length:
            if len(value) < self.min_length:
                raise ValidationError('{} is shorter than {} characters'.format(self.property_name, self.min_length))

        self.validator(value)

        return value


class Email(GraphProperty):
    """
    Email Data property type
    """
    data_type = "String"
    validator = validate_email

    def __init__(self, *args, **kwargs):
        self.encoding = kwargs.pop('encoding', 'utf-8')
        if 'default' in kwargs and isinstance(kwargs['default'], string_types):
            if not PY3:
                kwargs['default'] = kwargs['default'].encode(self.encoding)
        super(Email, self).__init__(*args, **kwargs)

    def validate(self, value):
        # Make sure it gets encoded correctly
        if isinstance(value, text_type) and not PY3:
            value = value.encode(self.encoding)

        value = super(Email, self).validate(value)

        # This should never happen unless the validator is changed
        if value in (None, [], (), {}):  # pragma: no cover
            return None

        if value is not None:
            if isinstance(value, (bytes, bytearray)) and not isinstance(value, str):
                return value.decode(self.encoding)

        # This should never happen unless the validator is changed
        if not isinstance(value, string_types):  # pragma: no cover
            raise ValidationError('%s is not a string' % type(value))

        self.validator(value)

        return value


class IPV4(GraphProperty):
    """
    IPv4 Data property type
    """
    data_type = "String"
    validator = validate_ipv4_address

    def __init__(self, *args, **kwargs):
        self.encoding = kwargs.pop('encoding', 'utf-8')
        if 'default' in kwargs and isinstance(kwargs['default'], string_types):
            if not PY3:
                kwargs['default'] = kwargs['default'].encode(self.encoding)
        super(IPV4, self).__init__(*args, **kwargs)

    def validate(self, value):
        # Make sure it gets encoded correctly
        if isinstance(value, text_type) and not PY3:
            value = value.encode(self.encoding)

        value = super(IPV4, self).validate(value)

        # This should never happen unless the validator is changed
        if value in (None, [], (), {}):  # pragma: no cover
            return None

        if value is not None:
            if isinstance(value, (bytes, bytearray)) and not isinstance(value, str):
                return value.decode(self.encoding)

        # This should never happen unless the validator is changed
        if not isinstance(value, string_types):  # pragma: no cover
            raise ValidationError('%s is not a string' % type(value))

        self.validator(value)

        return value


class IPV6(GraphProperty):
    """
    IPv6 Data property type
    """
    data_type = "String"
    validator = validate_ipv6_address

    def __init__(self, *args, **kwargs):
        self.encoding = kwargs.pop('encoding', 'utf-8')
        if 'default' in kwargs and isinstance(kwargs['default'], string_types):
            if not PY3:
                kwargs['default'] = kwargs['default'].encode(self.encoding)
        super(IPV6, self).__init__(*args, **kwargs)

    def validate(self, value):
        # Make sure it gets encoded correctly
        if isinstance(value, text_type) and not PY3:
            value = value.encode(self.encoding)

        value = super(IPV6, self).validate(value)

        if value is not None:
            if isinstance(value, (bytes, bytearray)) and not isinstance(value, str):
                return value.decode(self.encoding)

        # This shouldn't happend unless the validator is changed
        if value in (None, [], (), {}):  # pragma: no cover
            return None

        # This shouldn't happend unless the validator is changed
        if not isinstance(value, string_types):  # pragma: no cover
            raise ValidationError('%s is not a string' % type(value))

        self.validator(value)

        return value


class IPV6WithV4(GraphProperty):
    """
    IPv6 with Mapped/Translated/Embedded IPv4 Data property type
    """
    data_type = "String"
    validator = validate_ipv6_ipv4_address

    def __init__(self, *args, **kwargs):
        self.encoding = kwargs.pop('encoding', 'utf-8')
        if 'default' in kwargs and isinstance(kwargs['default'], string_types):
            if not PY3:
                kwargs['default'] = kwargs['default'].encode(self.encoding)
        super(IPV6WithV4, self).__init__(*args, **kwargs)

    def validate(self, value):
        # Make sure it gets encoded correctly
        if isinstance(value, text_type) and not PY3:
            value = value.encode(self.encoding)

        value = super(IPV6WithV4, self).validate(value)

        # This shouldn't happend unless the validator is changed
        if value in (None, [], (), {}):  # pragma: no cover
            return None

        if value is not None:
            if isinstance(value, (bytes, bytearray)) and not isinstance(value, str):
                return value.decode(self.encoding)

        # This shouldn't happend unless the validator is changed
        if not isinstance(value, string_types):  # pragma: no cover
            raise ValidationError('%s is not a string' % type(value))

        self.validator(value)

        return value


class Slug(GraphProperty):
    """
    Slug Data property type
    """
    data_type = "String"
    validator = validate_slug

    def __init__(self, *args, **kwargs):
        self.encoding = kwargs.pop('encoding', 'utf-8')
        if 'default' in kwargs and isinstance(kwargs['default'], string_types):
            if not PY3:
                kwargs['default'] = kwargs['default'].encode(self.encoding)
        super(Slug, self).__init__(*args, **kwargs)

    def validate(self, value):
        # Make sure it gets encoded correctly
        if isinstance(value, text_type) and not PY3:
            value = value.encode(self.encoding)

        value = super(Slug, self).validate(value)

        if value in (None, [], (), {}):
            return None

        if value is not None:
            if isinstance(value, (bytes, bytearray)) and not isinstance(value, str):
                return value.decode(self.encoding)

        if not isinstance(value, string_types):
            raise ValidationError('%s is not a string' % type(value))

        self.validator(value)

        return value



from __future__ import unicode_literals
import re
try:
    from urllib.parse import urlsplit, urlunsplit
except ImportError:     # Python 2
    from urlparse import urlsplit, urlunsplit

from mogwai._compat import string_types, text_type, float_types, integer_types, array_types, bool_types, print_
from mogwai.exceptions import MogwaiException, ValidationError
import datetime
from pytz import utc
from decimal import Decimal as _D
from collections import Iterable

# These values, if given to validate(), will trigger the self.required check.
EMPTY_VALUES = (None, '', [], (), {})


class BaseValidator(object):
    message = 'Enter a valid value.'
    code = 'invalid'

    def __init__(self, message=None, code=None):
        if message is not None:
            self.message = message
        if code is not None:
            self.code = code

    def __call__(self, value):
        """
        Validates that the input passes validation
        """
        return value

pass_all_validator = BaseValidator()


class BooleanValidator(BaseValidator):
    message = 'Enter a valid Boolean.'

    def __call__(self, value):
        if not isinstance(value, bool_types):
            raise ValidationError(self.message, self.code)
        return value

bool_validator = BooleanValidator()


class NumericValidator(BaseValidator):
    message = 'Enter a valid number.'
    data_types = float_types + integer_types + (_D, )

    def __call__(self, value):
        if not isinstance(value, self.__class__.data_types):
            raise ValidationError(self.message, code=self.code)
        return value

numeric_validator = NumericValidator()


class FloatValidator(NumericValidator):
    data_types = float_types


float_validator = FloatValidator()


class DecimalValidator(NumericValidator):
    data_types = float_types + (_D, )


decimal_validator = DecimalValidator()


class IntegerValidator(NumericValidator):
    data_types = integer_types


integer_validator = IntegerValidator()


class LongValidator(NumericValidator):
    data_types = integer_types


long_validator = LongValidator()


class PositiveIntegerValidator(NumericValidator):
    data_types = integer_types

    def __call__(self, value):
        super(PositiveIntegerValidator, self).__call__(value)
        if value < 0:
            raise ValidationError("Value must be 0 or greater")
        return value

positive_integer_validator = PositiveIntegerValidator()


class StringValidator(BaseValidator):
    message = 'Enter a valid string: {}'
    data_type = string_types

    def __call__(self, value):
        if not isinstance(value, self.data_type):
            raise ValidationError(self.message.format(value), code=self.code)
        return value


string_validator = StringValidator()


class ListValidator(BaseValidator):
    message = 'Enter a valid list'
    data_types = array_types

    def __call__(self, value):
        if not isinstance(value, self.data_types):
            raise ValidationError(self.message, code=self.code)
        return value

list_validator = ListValidator()


class DictValidator(BaseValidator):
    message = 'Enter a valid dict'

    def __call__(self, value):
        if not isinstance(value, dict):
            raise ValidationError(self.message, code=self.code)
        return value

dict_validator = DictValidator()


class DateTimeValidator(BaseValidator):
    message = 'Not a valid DateTime: {}'

    def __call__(self, value):
        if not isinstance(value, datetime.datetime) and value is not None:
            raise ValidationError(self.message.format(value), code=self.code)
        return value

datetime_validator = DateTimeValidator()


class DateTimeUTCValidator(BaseValidator):
    message = 'Not a valid UTC DateTime: {}'

    def __call__(self, value):
        super(DateTimeUTCValidator, self).__call__(value)
        if value is None:
            return
        if not isinstance(value, datetime.datetime) and value is not None:
            raise ValidationError(self.message.format(value), code=self.code)
        if value and value.tzinfo != utc:
            #print_("Got value with timezone: {} - {}".format(value, value.tzinfo))
            try:
                value = value.astimezone(tz=utc)
            except ValueError:  # last ditch effort
                try:
                    value = value.replace(tz=utc)
                except (AttributeError, TypeError):
                    raise ValidationError(self.message.format(value), code=self.code)
            except AttributeError:  # pragma: no cover
                # This should never happen, unless it isn't a datetime object
                raise ValidationError(self.message % (value, ), code=self.code)
        #print_("Datetime passed validation: {} - {}".format(value, value.tzinfo))
        return value


datetime_utc_validator = DateTimeUTCValidator()


class RegexValidator(BaseValidator):
    regex = ''

    def __init__(self, regex=None, message=None, code=None):
        super(RegexValidator, self).__init__(message=message, code=code)
        if regex is not None:
            self.regex = regex

        # Compile the regex if it was not passed pre-compiled.
        if isinstance(self.regex, string_types):  # pragma: no cover
            self.regex = re.compile(self.regex)

    def __call__(self, value):
        """
        Validates that the input matches the regular expression.
        """
        if not self.regex.search(text_type(value)):
            raise ValidationError(self.message, code=self.code)
        else:
            return value


class URLValidator(RegexValidator):
    regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|'  # ...or ipv4
        r'\[?[A-F0-9]*:[A-F0-9:]+\]?)'  # ...or ipv6
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    message = 'Enter a valid URL address: {}'
    code = 'invalid'

    def __call__(self, value):
        try:
            super(URLValidator, self).__call__(value)
        except ValidationError as e:
            # Trivial case failed. Try for possible IDN domain
            if value:
                value = text_type(value)
                scheme, netloc, path, query, fragment = urlsplit(value)
                try:
                    netloc = netloc.encode('idna').decode('ascii')  # IDN -> ACE
                except UnicodeError:  # invalid domain part
                    raise ValidationError(self.message.format(value), code=self.code)
                url = urlunsplit((scheme, netloc, path, query, fragment))
                return super(URLValidator, self).__call__(url)
            else:
                raise ValidationError(self.message.format(value), code=self.code)
        return value


class EmailValidator(RegexValidator):

    regex = re.compile(
        r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*"  # dot-atom
        # quoted-string, see also http://tools.ietf.org/html/rfc2822#section-3.2.5
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-\011\013\014\016-\177])*"'
        r')@((?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)$)'  # domain
        r'|\[(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\]$',  # literal form, ipv4 address (SMTP 4.1.3)
        re.IGNORECASE)

    message = 'Enter a valid email address: {}'
    code = 'invalid'

    def __call__(self, value):
        try:
            super(EmailValidator, self).__call__(value)
        except ValidationError as e:
            # Trivial case failed. Try for possible IDN domain-part
            if value and isinstance(value, string_types) and '@' in value:
                parts = value.split('@')
                try:
                    parts[-1] = parts[-1].encode('idna').decode('ascii')
                except UnicodeError:
                    raise ValidationError(self.message.format(value), code=self.code)
                super(EmailValidator, self).__call__('@'.join(parts))
            else:
                raise ValidationError(self.message.format(value), code=self.code)
        return value

validate_email = EmailValidator()

slug_re = re.compile(r'^[-a-zA-Z0-9_]+$')
validate_slug = RegexValidator(slug_re, "Enter a valid 'slug' consisting of letters, numbers, underscores or hyphens.", 'invalid')

## IPv4
ipv4_re = re.compile(r'^(25[0-5]|2[0-4]\d|[0-1]?\d?\d)(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}$')
validate_ipv4_address = RegexValidator(ipv4_re, 'Enter a valid IPv4 address.', 'invalid')

# IPv6
ipv6_re = re.compile(
    r'('
    r'([0-9A-F]{1,4}:){7,7}[0-9A-F]{1,4}|'          # 1:2:3:4:5:6:7:8
    r'([0-9A-F]{1,4}:){1,7}:|'                      # 1::                              1:2:3:4:5:6:7::
    r'([0-9A-F]{1,4}:){1,6}:[0-9A-F]{1,4}|'         # 1::8             1:2:3:4:5:6::8  1:2:3:4:5:6::8
    r'([0-9A-F]{1,4}:){1,5}(:[0-9A-F]{1,4}){1,2}|'  # 1::7:8           1:2:3:4:5::7:8  1:2:3:4:5::8
    r'([0-9A-F]{1,4}:){1,4}(:[0-9A-F]{1,4}){1,3}|'  # 1::6:7:8         1:2:3:4::6:7:8  1:2:3:4::8
    r'([0-9A-F]{1,4}:){1,3}(:[0-9A-F]{1,4}){1,4}|'  # 1::5:6:7:8       1:2:3::5:6:7:8  1:2:3::8
    r'([0-9A-F]{1,4}:){1,2}(:[0-9A-F]{1,4}){1,5}|'  # 1::4:5:6:7:8     1:2::4:5:6:7:8  1:2::8
    r'[0-9A-F]{1,4}:((:[0-9A-F]{1,4}){1,6})|'       # 1::3:4:5:6:7:8   1::3:4:5:6:7:8  1::8
    r':((:[0-9A-F]{1,4}){1,7}|:)'                   # ::2:3:4:5:6:7:8  ::2:3:4:5:6:7:8 ::8       ::
    r')', re.IGNORECASE)
validate_ipv6_address = RegexValidator(ipv6_re, 'Enter a valid IPv6 address.', 'invalid')


# IPv6/4
ipv64_re = re.compile(
    ipv6_re.pattern[:-1] + r'|' +
    r'::(ffff(:0{1,4}){0,1}:){0,1}'  # ::255.255.255.255   ::ffff:255.255.255.255  ::ffff:0:255.255.255.255
    r'((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]).){3,3}'  # ... (IPv4-mapped IPv6 addresses and IPv4-translated addresses)
    r'(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|'
    r'([0-9a-fA-F]{1,4}:){1,4}:'  # 2001:db8:3:4::192.0.2.33  64:ff9b::192.0.2.33 (IPv4-Embedded IPv6 Address)
    r'((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]).){3,3}'
    r'(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])'
    r')', re.IGNORECASE)

validate_ipv6_ipv4_address = RegexValidator(ipv64_re, 'Enter a valid IPv4, IPv6, '
                                                      '(IPv4-mapped, IPv4 Translated or IPv4-Embedded) IPv6 address',
                                            'invalid')


validate_url = URLValidator()

re_uuid = re.compile(r'[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}')
validate_uuid4 = RegexValidator(re_uuid, 'Enter a valid UUID4.', 'invalid')
validate_uuid1 = RegexValidator(re_uuid, 'Enter a valid UUID1.', 'invalid')



from __future__ import unicode_literals
from nose.plugins.attrib import attr
from mogwai.models import Vertex, Edge
from mogwai.properties import String, Integer
from mogwai.exceptions import MogwaiRelationshipException
from mogwai.tests.base import BaseMogwaiTestCase, TestEdgeModel, TestVertexModel, counter
from mogwai.relationships.base import Relationship


class TestVertex2Model(Vertex):
    element_type = 'test_vertex2_model'

    name = String(default='test_vertex')
    test_val = Integer(default=counter)


class TestEdge2Model(Edge):
    label = 'test_edge2_model'

    name = String(default='test_edge')
    test_val = Integer(default=counter)


@attr('unit', 'relationship')
class GraphRelationshipBaseTestCase(BaseMogwaiTestCase):
    """ Test Relationship Functionality """

    @classmethod
    def setUpClass(cls):
        super(GraphRelationshipBaseTestCase, cls).setUpClass()
        cls.relationship_base_cls = Relationship
        cls.edge_model = TestEdgeModel
        cls.vertex_model = TestVertexModel
        cls.vertex_start = TestVertexModel.create(name='test relationship')

    @classmethod
    def tearDownClass(cls):
        cls.vertex_start.delete()
        super(GraphRelationshipBaseTestCase, cls).tearDownClass()

    def test_instantiation(self):
        """ Test that the Relationship is properly Instantiated """

        # setup relationship
        relationship = self.relationship_base_cls(self.edge_model, self.vertex_model)
        self.assertIsNone(relationship.top_level_vertex_class)
        self.assertIsNone(relationship.top_level_vertex)

        with self.assertRaises(MogwaiRelationshipException):
            relationship.create({}, {})

    def test_relationship_io(self):
        """ Test Relationship GraphDB interaction for querying Edges and Vertices """


        # setup relationship
        relationship = self.relationship_base_cls(self.edge_model, self.vertex_model)
        relationship.top_level_vertex = self.vertex_start
        relationship.top_level_vertex_class = self.vertex_model

        vertices = relationship.vertices()
        self.assertEqual(len(vertices), 0)

        v2 = self.vertex_model.create(name='other side relationship')
        e1 = self.edge_model.create(v2, self.vertex_start)

        vertices = relationship.vertices()
        self.assertEqual(len(vertices), 1)

        edges = relationship.edges()
        self.assertEqual(len(edges), 1)

        e1.delete()
        v2.delete()

    def test_relationship_control(self):
        """ Test Relationship Constraint system """

        # setup relationship
        relationship = self.relationship_base_cls(self.edge_model, self.vertex_model)
        self.assertTrue(relationship.allowed(self.edge_model, self.vertex_model))

        class BadEdgeModel(object):
            pass

        class BadVertexModel(object):
            pass

        self.assertFalse(relationship.allowed(self.edge_model, BadVertexModel))
        self.assertFalse(relationship.allowed(BadEdgeModel, self.vertex_model))
        self.assertFalse(relationship.allowed(BadEdgeModel, BadVertexModel))

    def test_relationship_query(self):
        """ Test Relationship Query system

        The Query tests are in a separate tests, we only care that we get the same instantiated query class and
        functionally operates
        """

        # setup relationship
        relationship = self.relationship_base_cls(self.edge_model, self.vertex_model)
        relationship.top_level_vertex = self.vertex_start
        relationship.top_level_vertex_class = self.vertex_model

        v2 = self.vertex_model.create(name='other side relationship')
        e1 = self.edge_model.create(v2, self.vertex_start)

        from mogwai.models import Query, IN

        # default query
        query = relationship.query()
        self.assertIsInstance(query, Query)
        result = query.direction(IN)._get_partial()
        self.assertEqual(result, "g.v(id).query().labels('test_edge_model').direction(IN)")

        # specified edge_types query
        query = relationship.query(edge_types=self.edge_model)
        self.assertIsInstance(query, Query)
        result = query.direction(IN)._get_partial()
        self.assertEqual(result, "g.v(id).query().labels('test_edge_model').direction(IN)")

        e1.delete()
        v2.delete()

    def test_relationship_creation(self):
        """ Test Relationship Vertex and Edge Creation mechanism """

        # setup relationship
        relationship = self.relationship_base_cls(TestEdge2Model, TestVertex2Model)
        relationship.top_level_vertex = self.vertex_start
        relationship.top_level_vertex_class = self.vertex_model

        e1, v2 = relationship.create(edge_params={}, vertex_params={'name': 'other side relationship'})
        self.assertIsInstance(e1, Edge)
        self.assertIsInstance(v2, Vertex)

        vertex_result = relationship.vertices()
        self.assertEqual(len(vertex_result), 1)
        self.assertEqual(vertex_result[0], v2)

        edge_result = relationship.edges()
        self.assertEqual(len(edge_result), 1)
        self.assertEqual(edge_result[0], e1)

        e1.delete()
        v2.delete()


## TODO: Reference: Tooling from werkzeug tools - citation needed
from __future__ import unicode_literals
import sys
from rexpro.exceptions import RexProConnectionException, RexProScriptException
from mogwai._compat import reraise, string_types, PY2
from mogwai import connection
from mogwai.exceptions import MogwaiBlueprintsWrapperException
from factory import base
from rexpro.connectors.sync import RexProSyncConnection


class ImportStringError(ImportError):
    """Provides information about a failed :func:`import_string` attempt."""

    # String in dotted notation that failed to be imported.
    import_name = None
    # Wrapped exception.
    exception = None

    def __init__(self, import_name, exception):
        self.import_name = import_name
        self.exception = exception

        msg = (
            'import_string() failed for %r. Possible reasons are:\n\n'
            '- missing __init__.py in a package;\n'
            '- package or module path not included in sys.path;\n'
            '- duplicated package or module name taking precedence in '
            'sys.path;\n'
            '- missing module, class, function or variable;\n\n'
            'Debugged import:\n\n%s\n\n'
            'Original exception:\n\n%s: %s')

        name = ''
        tracked = []
        for part in import_name.replace(':', '.').split('.'):
            name += (name and '.') + part
            imported = import_string(name, silent=True)
            if imported:
                tracked.append((name, getattr(imported, '__file__', None)))
            else:
                track = ['- %r found in %r.' % (n, i) for n, i in tracked]
                track.append('- %r not found.' % name)
                msg = msg % (import_name, '\n'.join(track),
                             exception.__class__.__name__, str(exception))
                break

        ImportError.__init__(self, msg)

    def __repr__(self):  # pragma: no cover
        return '<%s(%r, %r)>' % (self.__class__.__name__, self.import_name,
                                 self.exception)


def import_string(import_name, silent=False):
    """Imports an object based on a string.  This is useful if you want to
    use import paths as endpoints or something similar.  An import path can
    be specified either in dotted notation (``xml.sax.saxutils.escape``)
    or with a colon as object delimiter (``xml.sax.saxutils:escape``).

    If `silent` is True the return value will be `None` if the import fails.

    :param import_name: the dotted name for the object to import.
    :param silent: if set to `True` import errors are ignored and
                   `None` is returned instead.
    :return: imported object
    """
    assert isinstance(import_name, string_types)
    # force the import name to automatically convert to strings
    import_name = str(import_name)
    try:
        if ':' in import_name:
            module, obj = import_name.split(':', 1)
        elif '.' in import_name:
            module, obj = import_name.rsplit('.', 1)
        else:
            return __import__(import_name)
        # __import__ is not able to handle unicode setup_strings in the fromlist
        # if the module is a package
        if PY2 and isinstance(obj, unicode):  # pragma: no cover
            obj = obj.encode('utf-8')
        try:
            return getattr(__import__(module, None, None, [obj]), obj)
        except (ImportError, AttributeError):
            # support importing modules not yet set up by the parent module
            # (or package for that matter)
            modname = module + '.' + obj
            __import__(modname)
            return sys.modules[modname]
    except ImportError as e:
        if not silent:
            reraise(
                ImportStringError,
                ImportStringError(import_name, e),
                sys.exc_info()[2])


class _Missing(object):

    def __repr__(self):  # pragma: no cover
        return 'no value'

    def __reduce__(self):  # pragma: no cover
        return '_missing'

_missing = _Missing()


class cached_property(object):
    """A decorator that converts a function into a lazy property.  The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value::

        class Foo(object):

            @cached_property
            def foo(self):
                # calculate something important here
                return 42

    The class has to have a `__dict__` in order for this property to
    work.
    """

    # implementation detail: this property is implemented as non-data
    # descriptor.  non-data descriptors are only invoked if there is
    # no entry with the same name in the instance's __dict__.
    # this allows us to completely get rid of the access function call
    # overhead.  If one choses to invoke __get__ by hand the property
    # will still work as expected because the lookup logic is replicated
    # in __get__ for manual invocation.

    def __init__(self, func, name=None, doc=None):
        self.__name__ = name or func.__name__
        self.__module__ = func.__module__
        self.__doc__ = doc or func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.__name__, _missing)
        if value is _missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value


class LazyImportClass(object):
    """ Lazily load and return a class """

    def __init__(self, import_name):
        try:
            self.__module__, self.__name__ = import_name.split('.', 1)
        except ValueError:  # pragma: no cover
            self.__module__ = ''
            self.__name__ = import_name
        self.__module__ = self.__module__ .lstrip('.')
        self.import_name = import_name.lstrip('.')
        #self.__setup_instantiated_vertex = getattr(self.klass, '__setup_instantiated_vertex')

    @cached_property
    def klass(self):
        """ Imports the class and caches """
        return import_string(self.import_name)

    def __call__(self, *args, **kwargs):
        return self.klass(*args, **kwargs)

    def __setup_instantiated_vertex(self, *args, **kwargs):  # pragma: no cover
        return getattr(self.klass, '__setup_instantiated_vertex')(*args, **kwargs)


class Factory(base.Factory):
    """Factory for Mogwai models. """

    ABSTRACT_FACTORY = True

    _associated_model = None

    @classmethod
    def _load_target_class(cls, class_definition):
        """ So we can support potential circular import problems, by using normal import_string import specification.
        """
        associated_class = super(Factory, cls)._load_target_class(class_definition)

        if isinstance(associated_class, string_types) and '.' in associated_class:
            if cls._associated_model is None:
                cls._associated_model = import_string(associated_class)
                cls._associated_model.FACTORY_CLASS = cls
            return cls._associated_model

        if associated_class and associated_class.FACTORY_CLASS is None:
            associated_class.FACTORY_CLASS = cls

        return associated_class

    @classmethod
    def _create(cls, target_class, *args, **kwargs):
        """Create an instance of the model, and save it to the database."""
        from mogwai.models import Edge
        if issubclass(target_class.__class__, Edge):
            assert ('outV' in kwargs and 'inV' in kwargs), "Edges require in and out Vertices"
            obj = target_class.create(outV=kwargs.get('outV'), inV=kwargs.get('inV'), *args, **kwargs)
        else:
            obj = target_class.create(*args, **kwargs)
        return obj


class SessionPoolManager(object):
    """A context manager that exposes a pool whose connections share the same session.
    If RexPro is used without concurrency, it is not necessary to use connections
    from the pool explicitly. `connection.execute_query` is patched in this case
    to use connections from the pool by default. In concurrent mode the pool always
    needs to be used explicitly, because we cannot rely on global context for patching.

    :param dict bindings: variables that are available throughout the session (optional)
    :param int pool_size: the maximum number of simultaneous connections for this pool
    :return: the used RexPro connection pool (with default session)
    """

    def __init__(self, bindings=None, pool_size=10):
        self.bindings = bindings
        self.pool_size = pool_size

    def __enter__(self):
        # assign original execute_query()
        self.original_execute_query = connection.execute_query
        # create a pool with session
        self.pool = connection.CONNECTION_POOL_TYPE(pool_size=self.pool_size, with_session=True,
                                                    **connection.HOST_PARAMS)

        # assign optional binding variables
        if self.bindings:
            connection.execute_query('g', params=self.bindings, isolate=False, pool=self.pool)

        # patch execute_query if we're running non-concurrently
        if connection.CONNECTION_TYPE == RexProSyncConnection:
            # shadow execute_query with default self.pool
            def execute_in_pool(query, params=None, transaction=True, isolate=True,
                                pool=self.pool, *args, **kwargs):
                params = params or {}
                return self.original_execute_query(
                    query, params=params, transaction=transaction, isolate=isolate, pool=pool, *args, **kwargs
                )

            # patch execute_query to re-use the pool with session
            connection.execute_query = execute_in_pool

        return self.pool

    def __exit__(self, exc, message, traceback):
        # replace original execute_query()
        connection.execute_query = self.original_execute_query
        # commit and close all connections
        self.pool.close_all(force_commit=True)

        return False


class BlueprintsWrapper(SessionPoolManager):
    """Abstract implementation for using a Blueprints graph wrapper
    within a persisted transaction. Within this transaction `g`
    refers to the wrapped graph.

    :param str class_name: the Blueprints Implementation class name
    :param str setup: an iterable with additional groovy statements that are
        executed upon initialization. (optional, default: [])
    :return: the used RexPro connection pool (with default session)
    :raises MogwaiBlueprintsWrapperException: if no class name is provided
    """

    def __init__(self, class_name=None, setup=None, *args, **kwargs):
        super(BlueprintsWrapper, self).__init__(*args, **kwargs)

        if not class_name: # pragma: no cover
            raise MogwaiBlueprintsWrapperException("A wrapper class name is required.")
        self.g_assignment = "g = new {}(g)".format(class_name)
        self.setup = setup or []

    def __enter__(self):
        pool = super(BlueprintsWrapper, self).__enter__()

        # execute g_assignment
        resp = connection.execute_query(self.g_assignment, transaction=False, isolate=False, pool=pool)

        # provide a dummy stopTransaction() on non-transactional graphs
        connection.execute_query(
            "if (!g.metaClass.respondsTo(g, 'stopTransaction')) { g.metaClass.stopTransaction = {null} }",
            transaction=False, isolate=False, pool=self.pool
        )

        # execute wrapper setup
        for statement in self.setup:
            connection.execute_query(statement, transaction=False, isolate=False, pool=pool)

        return pool

    def __exit__(self, exc, message, traceback):
        # execute g = g.baseGraph
        connection.execute_query("g = g.baseGraph", transaction=False, isolate=False, pool=self.pool)

        return super(BlueprintsWrapper, self).__exit__(exc, message, traceback)


class PartitionGraph(BlueprintsWrapper):
    """Wrapper for the Blueprints PartitionGraph, which is
    documented by https://github.com/tinkerpop/blueprints/wiki/Partition-Implementation

    :param str write: the default read+write partition.
    :param iterable read: optional read partitions.
    :return: the used RexPro connection pool (with default session)
    :raises MogwaiBlueprintsWrapperException: if no write partition is provided
    """

    def __init__(self, write=None, read=None, *args, **kwargs):
        if not write: # pragma: no cover
            raise MogwaiBlueprintsWrapperException("A write partition is required.")
        read = read or []
        super(PartitionGraph, self).__init__('PartitionGraph', *args, **kwargs)
        self.g_assignment = "g = new PartitionGraph(g, '_partition', '{}')".format(write)
        self.setup = ["g.addReadPartition('{}')".format(rp) for rp in read]


import collections
import pyparsing
import re


# Cache of parsed files
_parsed_file_cache = {}
GroovyImport = collections.namedtuple('GroovyImport', ['comment_list', 'import_strings', 'import_list'])
GroovyFunction = collections.namedtuple('GroovyFunction', ['name', 'args', 'body', 'defn'])
GroovyFileDef = collections.namedtuple('GroovyFileDefinition', ['functions', 'imports', 'filename'])


class GroovyFunctionParser(object):
    """
    Given a string containing a single function definition this class will parse the function definition and return
    information regarding it.
    """

    # Simple Groovy sub-grammar definitions
    KeywordDef = pyparsing.Keyword('def')
    VarName = pyparsing.Regex(r'[A-Za-z_]\w*')
    FuncName = VarName
    FuncDefn = KeywordDef + FuncName + "(" + pyparsing.delimitedList(VarName) + ")" + "{"

    @classmethod
    def parse(cls, data):
        """
        Parse the given function definition and return information regarding the contained definition.

        :param data: The function definition in a string
        :type data: str | basestring
        :rtype: dict

        """
        try:
            # Parse the function here
            result = cls.FuncDefn.parseString(data)
            result_list = result.asList()
            args = result_list[3:result_list.index(')')]
            # Return single line or multi-line function body
            fn_body = re.sub(r'[^\{]+\{', '', data, count=1)
            parts = fn_body.strip().split('\n')
            fn_body = '\n'.join(parts[0:-1])
            return GroovyFunction(result[1], args, fn_body, data)
        except Exception as ex:
            return None


class GroovyImportParser(object):
    """
    Given a string containing a single import definition this class will parse the import definition and return
    information regarding it.
    """

    # Simple Groovy sub-grammar definitions
    ImportDef = pyparsing.Suppress(pyparsing.Keyword('import'))
    ImportVarName = pyparsing.Regex(r'[A-Za-z_.\*]*')
    CommentVar = pyparsing.Word(pyparsing.alphas, pyparsing.alphanums).setName('comment')
    OptionalSpace = pyparsing.Optional(' ')
    ImportDefn = ImportDef + \
                 pyparsing.delimitedList(ImportVarName, delim='.').setResultsName('imports') + \
                 pyparsing.Suppress(";") + \
                 pyparsing.Optional(
                     pyparsing.Suppress('//') +
                     pyparsing.delimitedList(CommentVar, delim=pyparsing.Empty()).setResultsName('comment')
                 )

    @classmethod
    def parse(cls, data):
        """
        Parse the given import and return information regarding the contained import statement.

        :param data: The import statement in a string
        :type data: str | basestring
        :rtype: dict

        """
        try:
            # Parse the function here
            result = cls.ImportDefn.parseString(data)
            package_list = []
            if 'imports' in result:
                package_list = result['imports'].asList()

            comment_list = []
            if 'comment' in result:
                comment_list = result['comment'].asList()

            return GroovyImport(comment_list,
                                package_list,
                                ['import {};'.format(package) for package in package_list])
        except Exception as ex:
            return None


def parse(filename):
    """
    Parse Groovy code in the given file and return a list of information about each function necessary for usage in
    queries to database.

    :param filename: The file containing groovy code.
    :type filename: str
    :rtype: list

    """
    # Check cache before parsing file
    global _parsed_file_cache
    if filename in _parsed_file_cache:
        return _parsed_file_cache[filename]

    ImportDefnRegexp = r'^import.*'
    FuncDefnRegexp = r'^def.*\{'
    FuncEndRegexp = r'^\}.*$'
    passedFirstFunction = False
    with open(filename, 'r') as f:
        file_lines = [line.rstrip('\n') for line in f.readlines()]
    all_fns = []
    all_imports = []
    fn_lines = ''
    for line in file_lines:
        if not passedFirstFunction and re.match(ImportDefnRegexp, line):
            all_imports.append(line)
        elif len(fn_lines) > 0:
            if re.match(FuncEndRegexp, line):
                fn_lines += line + "\n"
                all_fns.append(fn_lines)
                fn_lines = ''
            else:
                fn_lines += line + "\n"
        elif re.match(FuncDefnRegexp, line):
            fn_lines += line + "\n"
            passedFirstFunction = True

    import_results = [GroovyImportParser.parse(im) for im in all_imports]
    func_results = [GroovyFunctionParser.parse(fn) for fn in all_fns]

    result = GroovyFileDef(func_results, import_results, filename)
    _parsed_file_cache[filename] = result
    return result



from __future__ import unicode_literals


class MogwaiException(Exception):
    """ Generic Base Exception for Mogwai Library """
    pass


class MogwaiConnectionError(MogwaiException):
    """ Problem connecting with Titan """
    pass


class MogwaiGraphMissingError(MogwaiException):
    """ Graph with specified name does not exist """
    pass


class MogwaiQueryError(MogwaiException):
    """ Exception thrown when a query error occurs """
    pass


class ValidationError(MogwaiException):
    """ Exception thrown when a property value validation error occurs """

    def __init__(self, *args, **kwargs):
        self.code = kwargs.pop('code', None)
        super(MogwaiException, self).__init__(*args, **kwargs)


class ElementDefinitionException(MogwaiException):
    """ Error in element definition """
    pass


class ModelException(MogwaiException):
    """ Error in model """
    pass


class SaveStrategyException(MogwaiException):
    """ Exception thrown when a Save Strategy error occurs """
    pass


class MogwaiGremlinException(MogwaiException):
    """ Exception thrown when a Gremlin error occurs """
    pass


class MogwaiRelationshipException(MogwaiException):
    """ Exception thrown when a Relationship error occurs """
    pass


class MogwaiMetricsException(MogwaiException):
    """ Exception thrown when a metric system error occurs """
    pass


class MogwaiBlueprintsWrapperException(MogwaiException):
    """ Exception thrown when a Blueprints wrapper error occurs """
    pass



from __future__ import unicode_literals
from nose.plugins.attrib import attr
from .base_tests import GraphPropertyBaseClassTestCase
from mogwai.properties.properties import Float, Double
from mogwai.models import Vertex
from mogwai._compat import print_


@attr('unit', 'property', 'property_float')
class FloatPropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = Float
    good_cases = (1.1, 0.0, None)
    bad_cases = ('', 'a', 1, [], [1], {}, {'a': 1})


class FloatTestVertex(Vertex):
    element_type = 'test_float_vertex'

    test_val = Float()


@attr('unit', 'property', 'property_float')
class FloatVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_float_io(self):
        print_("creating vertex")
        dt = FloatTestVertex.create(test_val=1.1)
        print_("getting vertex from vertex: %s" % dt)
        dt2 = FloatTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = FloatTestVertex.create(test_val=2.2)
        print_("\ncreated vertex: %s" % dt)
        dt2 = FloatTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, 2.2)
        print_("deleting vertex")
        dt2.delete()

@attr('unit', 'property', 'property_double')
class DoublePropertyTestCase(FloatPropertyTestCase):
    klass = Double


class DoubleTestVertex(Vertex):
    element_type = 'test_double_vertex'

    test_val = Double()


@attr('unit', 'property', 'property_double')
class DoubleVertexTestCase(GraphPropertyBaseClassTestCase):

    def test_double_io(self):
        print_("creating vertex")
        dt = DoubleTestVertex.create(test_val=1.1)
        print_("getting vertex from vertex: %s" % dt)
        dt2 = DoubleTestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = DoubleTestVertex.create(test_val=2.2)
        print_("\ncreated vertex: %s" % dt)
        dt2 = DoubleTestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, 2.2)
        print_("deleting vertex")
        dt2.delete()


from __future__ import unicode_literals
import logging
import warnings
from functools import wraps

from mogwai._compat import array_types, string_types
from mogwai.tools import LazyImportClass
from mogwai.exceptions import MogwaiRelationshipException

from mogwai.constants import IN, OUT, BOTH

logger = logging.getLogger(__name__)


def requires_vertex(method):
    @wraps(method)
    def method_wrapper(self, *args, **kwargs):
        if self.top_level_vertex:
            return method(self, *args, **kwargs)
        else:
            raise MogwaiRelationshipException("No Vertex Instantiated")
    return method_wrapper


class Relationship(object):

    """  Define incoming and outgoing relationships that exist. Also enforce schema IN, OUT and BOTH directions

    Warn if queries return schema violations.
    """

    def __init__(self, edge_class, vertex_class, direction=BOTH, strict=True, gremlin_path=None, vertex_callback=None,
                 edge_callback=None, query_callback=None, create_callback=None):
        from mogwai.models import Edge, Vertex
        self.edge_classes = self.__create_class_tuple(edge_class, enforce_type=Edge)
        self.vertex_classes = self.__create_class_tuple(vertex_class, enforce_type=Vertex)
        assert direction in (IN, OUT, BOTH), \
            "Direction of Relationship must be of one in (%s, %s, %s)" % (IN, OUT, BOTH)
        self.direction = direction
        self.strict = strict
        if gremlin_path:  # pragma: no cover
            self.gremlin_path = gremlin_path
        self.vertex_callback = vertex_callback
        self.edge_callback = edge_callback
        self.query_callback = query_callback
        self.create_callback = create_callback
        self.top_level_vertex_class = None
        self.top_level_vertex = None

    def _setup_instantiated_vertex(self, vertex):
        self.top_level_vertex = vertex
        self.top_level_vertex_class = vertex.__class__

    def __create_class_tuple(self, model_class, enforce_type=None):
        """ Take in an string, array of classes, or a single class and make a tuple of said referenced classes

        :param model_class: Input to be transformed into reference class(es)
        :type model_class: string_types | array_types | mogwai.models.Edge | mogwai.models.Vertex
        :param enforce_type: Enforce a specific model type? If not provided, everything that is resolved passes,
                             otherwise if a type is given, the classes are filtered out that don't match.
        :type enforce_type: None | mogwai.models.Vertex | mogwai.models.Edge
        :rtype: tuple[enforce_type | Object]
        """
        if isinstance(model_class, string_types):
            model_classes = (LazyImportClass(model_class), )
        elif isinstance(model_class, array_types):
            model_classes = []
            for mc in model_class:
                if isinstance(mc, string_types):
                    model_classes.append(LazyImportClass(mc))
                else:
                    model_classes.append(mc)
            model_classes = tuple(model_classes)
        else:
            model_classes = (model_class, )

        if not enforce_type:
            return model_classes
        else:
            final_classes = []
            for kls in model_classes:
                if isinstance(kls, LazyImportClass) or issubclass(kls, enforce_type):
                    final_classes.append(kls)
                else:
                    warnings.warn("Relationship constraint is not derived from %s and will be ignored!" % enforce_type,
                                  category=SyntaxWarning)
            return tuple(final_classes)

    @requires_vertex
    def vertices(self, limit=None, offset=None, callback=None):
        """ Query and return all Vertices attached to the current Vertex

        TODO: fix this, the instance method isn't properly setup
        :param limit: Limit the number of returned results
        :type limit: int | long
        :param offset: Query offset of the number of paginated results
        :type offset: int | long
        :param callback: (Optional) Callback function to handle results
        :type callback: method
        :rtype: List[mogwai.models.Vertex] | Object
        """
        allowed_elts = []
        allowed_vlts = []
        for e in self.edge_classes:
            allowed_elts += [e.get_label()]
        for v in self.vertex_classes:
            allowed_vlts += [v.get_element_type()]

        if limit is not None and offset is not None:
            start = offset
            end = offset + limit
        else:
            start = end = None

        operation = self.direction.lower() + 'V'

        result = getattr(self.top_level_vertex, operation)(*allowed_elts)
        if callback:
            return callback(result)
        elif self.vertex_callback:
            return self.vertex_callback(result)
        else:
            return result

    @requires_vertex
    def edges(self, limit=None, offset=None, callback=None):
        """ Query and return all Edges attached to the current Vertex

        TODO: fix this, the instance method isn't properly setup
        :param limit: Limit the number of returned results
        :type limit: int | long
        :param offset: Query offset of the number of paginated results
        :type offset: int | long
        :param callback: (Optional) Callback function to handle results
        :type callback: method
        :rtype: List[mogwai.models.Edge] | Object
        """
        allowed_elts = []
        for e in self.edge_classes:
            allowed_elts += [e.get_label()]

        if limit is not None and offset is not None:
            start = offset
            end = offset + limit
        else:
            start = end = None

        operation = self.direction.lower() + 'E'

        result = getattr(self.top_level_vertex, operation)(*allowed_elts)
        if callback:
            return callback(result)
        elif self.edge_callback:
            return self.edge_callback(result)
        else:
            return result

    def allowed(self, edge_type, vertex_type):
        """ Check whether or not the allowed Edge and Vertex type are compatible with the schema defined

        :param edge_type: Edge Class
        :type: mogwai.models.Edge
        :param vertex_type: Vertex Class
        :type: mogwai.models.Vertex
        :rtype: bool
        """
        if self.strict:
            if edge_type in self.edge_classes and vertex_type in self.vertex_classes:
                return True
            else:
                return False
        else:
            return True

    @requires_vertex
    def query(self, edge_types=None, callback=None):
        """ Generic Query method for quick access

        :param edge_types: List of Edge classes to query against
        :type edge_types: List[mogwai.models.Edge] | None
        :param callback: (Optional) Callback function to handle results
        :type callback: method
        :rtype: mogwai.models.query.Query | Object
        """
        #if not self.top_level_vertex:
        #    raise MogwaiRelationshipException("No vertex known to start with, this is an error")
        if edge_types:
            if not isinstance(edge_types, array_types):
                edge_types = (edge_types, )
            for et in edge_types:
                if not et in self.edge_classes:
                    raise MogwaiRelationshipException("Not a recognized edge label type, invalid schema")
        else:
            edge_types = self.edge_classes
        from mogwai.models.query import Query
        query = Query(self.top_level_vertex).labels(*edge_types)
        if callback:
            return callback(query)
        elif self.query_callback:
            return self.query_callback(query)
        else:
            return query

    def _create_entity(self, model_cls, model_params, outV=None, inV=None):
        """ Create Vertex and Edge between current Vertex and New Vertex

        :param model_cls: Vertex or Edge Class for the relationship
        :type model_cls: mogwai.models.Vertex | mogwai.models.Edge
        :param model_params: Vertex or Edge class parameters for instantiating the model
        :type model_params: dict
        :param outV: Outgoing Vertex if creating an Edge between two vertices (otherwise ignored)
        :type outV: mogwai.models.Vertex
        :param inV: Incoming Vertex if creating an Edge between two vertices (otherwise ignored)
        :type inV: mogwai.models.Vertex
        :rtype: mogwai.models.Vertex | mogwai.models.Edge
        """
        create_cls = model_cls._get_factory()

        from mogwai.models.edge import Edge
        if issubclass(model_cls, Edge):
            return create_cls(outV=outV, inV=inV, **model_params)
        else:
            return create_cls(**model_params)

    @requires_vertex
    def create(self, edge_params={}, vertex_params={}, edge_type=None, vertex_type=None, callback=None):
        """ Creates a Relationship defined by the schema

        :param edge_params: (Optional) Parameters passed to the instantiation method of the Edge
        :type edge_params: dict
        :param vertex_params: (Optional) Parameters passed to the instantiation method
        :type vertex_params: dict
        :param edge_type: (Optional) Edge class type, otherwise it defaults to the first Edge type known
        :type edge_type: mogwai.models.Edge | None
        :param edge_type: (Optional) Vertex class type, otherwise it defaults to the first Vertex type known
        :type edge_type: mogwai.models.Vertex | None
        :param callback: (Optional) Callback function to handle results
        :type callback: method
        :rtype: tuple(mogwai.models.Edge, mogwai.models.Vertex) | Object
        """
        #if not self.top_level_vertex:
        #    raise MogwaiRelationshipException("No existing vertex known, have you created a vertex?")
        if not vertex_type:
            vertex_type = self.vertex_classes[0]
        if not edge_type:
            edge_type = self.edge_classes[0]
        if not self.allowed(edge_type, vertex_type):
            raise MogwaiRelationshipException("That is not a valid relationship setup: %s <-%s-> %s" %
                                                (edge_type, self.direction, vertex_type))

        new_vertex = self._create_entity(vertex_type, vertex_params)
        if self.direction == IN:
            outV = new_vertex
            inV = self.top_level_vertex
        else:
            outV = self.top_level_vertex
            inV = new_vertex

        new_edge = self._create_entity(edge_type, edge_params, outV=outV, inV=inV)
        if callback:
            return callback(new_edge, new_vertex)
        elif self.create_callback:
            return self.create_callback(new_edge, new_vertex)
        else:
            return new_edge, new_vertex


from __future__ import unicode_literals
from nose.plugins.attrib import attr
from .base_tests import GraphPropertyBaseClassTestCase
from mogwai._compat import PY2
from mogwai.properties.properties import UUID
from mogwai.properties.base import GraphProperty, ValidationError
from mogwai.models import Vertex
from mogwai._compat import print_


@attr('unit', 'property', 'property_uuid')
class UUID4PropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = UUID
    good_cases = ('bb19eaed-c946-4cef-8001-7cc3357cc439', None)
    if PY2:
        bad_cases = ('val', [], (), {}, 0, long(1), False, 1.1)
    else:
        bad_cases = ('val', [], (), {}, 0, False, 1.1)

    def test_to_database_method(self):
        u = self.klass(default=None)
        self.assertIsNone(u.to_database(None))


class UUID4TestVertex(Vertex):
    element_type = 'test_uuid4_vertex'

    test_val = UUID()


@attr('unit', 'property', 'property_uuid')
class UUID4VertexTestCase(GraphPropertyBaseClassTestCase):

    def test_uuid4_io(self):
        print_("creating vertex")
        dt = UUID4TestVertex.create(test_val='bb19eaed-c946-4cef-8001-7cc3357cc439')
        print_("getting vertex from vertex: %s" % dt)
        dt2 = UUID4TestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = UUID4TestVertex.create(test_val='bb19eaed-c946-4cef-8001-7cc3357cc439')
        print_("\ncreated vertex: %s" % dt)
        dt2 = UUID4TestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, 'bb19eaed-c946-4cef-8001-7cc3357cc439')
        print_("deleting vertex")
        dt2.delete()


@attr('unit', 'property', 'property_uuid')
class UUID1PropertyTestCase(GraphPropertyBaseClassTestCase):
    klass = UUID(version='uuid1')
    good_cases = ('bb19eaed-c946-4cef-8001-7cc3357cc439', None)
    if PY2:
        bad_cases = ('val', [], (), {}, 0, long(1), False, 1.1)
    else:
        bad_cases = ('val', [], (), {}, 0, False, 1.1)

    def test_subclass(self):
        """ Test if Property is a GraphProperty """
        self.assertIsSubclass(self.klass.__class__, GraphProperty)

    def test_validation(self):
        for case in self.good_cases:
            print_("testing good case: %s" % (case, ))
            self.assertNotRaise(self.klass.validate, case)

        for case in self.bad_cases:
            print_("testing bad case: %s" % (case, ))
            self.assertRaises(ValidationError, self.klass.validate, case)


class UUID1TestVertex(Vertex):
    element_type = 'test_uuid1_vertex'

    test_val = UUID(version='uuid1')


@attr('unit', 'property', 'property_uuid')
class UUID1VertexTestCase(GraphPropertyBaseClassTestCase):

    def test_uuid1_io(self):
        print_("creating vertex")
        dt = UUID1TestVertex.create(test_val='bb19eaed-c946-4cef-8001-7cc3357cc439')
        print_("getting vertex from vertex: %s" % dt)
        dt2 = UUID1TestVertex.get(dt._id)
        print_("got vertex: %s\n" % dt2)
        self.assertEqual(dt2.test_val, dt.test_val)
        print_("deleting vertex")
        dt2.delete()

        dt = UUID1TestVertex.create(test_val='bb19eaed-c946-4cef-8001-7cc3357cc439')
        print_("\ncreated vertex: %s" % dt)
        dt2 = UUID1TestVertex.get(dt._id)
        print_("Got vertex: %s" % dt2)
        self.assertEqual(dt2.test_val, 'bb19eaed-c946-4cef-8001-7cc3357cc439')
        print_("deleting vertex")
        dt2.delete()


from __future__ import unicode_literals
from mogwai._compat import string_types, PY2
from nose.plugins.attrib import attr
from mogwai.tests import BaseMogwaiTestCase
from mogwai.models import Vertex, Edge
from mogwai import properties
from mogwai.tools import ImportStringError, import_string, cached_property, LazyImportClass, _Missing, Factory
import factory
from nose.tools import raises, nottest
from mogwai import connection
from mogwai.exceptions import MogwaiQueryError
from rexpro.exceptions import RexProScriptException
from mogwai.tools import BlueprintsWrapper, PartitionGraph
import eventlet


@attr('unit', 'tools')
class TestImportString(BaseMogwaiTestCase):

    def test_raises_import_string_error(self):
        with self.assertRaises(ImportStringError):
            import_string('does.not.exist')
        with self.assertRaises(ImportStringError):
            import_string('does.not:exist')

    def test_imports_correctly(self):
        cp = import_string('mogwai.tools:cached_property')
        self.assertIs(cp, cached_property)

    def test_import_string_error_class(self):
        ise = ImportStringError('mogwai.tools.cached_property', 'Should work')


class LazyImportObject(object):
    pass


@attr('unit', 'tools')
class TestLazyImportString(BaseMogwaiTestCase):

    def test_lazy_import(self):
        li = LazyImportClass('mogwai.tools._Missing')
        self.assertIsInstance(li, LazyImportClass)
        self.assertIs(li.klass, _Missing)
        self.assertIsInstance(li(), _Missing)


class TestFactoryVertex(Vertex):
    element_type = 'test_factory_vertex'

    name = properties.String(required=True, max_length=128)


class TestFactoryVertexFactory(Factory):
    FACTORY_FOR = TestFactoryVertex

    name = factory.Sequence(lambda n: 'Test Factory Vertex {}'.format(n))


class TestFactoryEdge(Edge):
    label = 'test_factory_edge'

    name = properties.String(required=True, max_length=128)


class TestFactoryEdgeFactory(Factory):
    FACTORY_FOR = TestFactoryEdge

    name = factory.Sequence(lambda n: 'Test Factory Edge {}'.format(n))


@attr('unit', 'tools', 'factory')
class TestModelFactory(BaseMogwaiTestCase):

    def test_vertex_factory(self):
        v1 = TestFactoryVertexFactory()
        self.assertIsInstance(v1, TestFactoryVertex)
        self.assertTrue(v1.name.startswith('Test Factory Vertex'))
        v1.delete()
        v2 = TestFactoryVertexFactory.create()
        self.assertIsInstance(v2, TestFactoryVertex)
        self.assertTrue(v2.name.startswith('Test Factory Vertex'))
        v2.delete()

    def test_edge_factory(self):
        v1 = TestFactoryVertexFactory()
        v2 = TestFactoryVertexFactory()
        e1 = TestFactoryEdgeFactory(outV=v1, inV=v2)
        e2 = TestFactoryEdgeFactory.create(outV=v1, inV=v2)
        self.assertIsInstance(e1, TestFactoryEdge)
        self.assertTrue(e1.name.startswith('Test Factory Edge'))
        self.assertIsInstance(e2, TestFactoryEdge)
        self.assertTrue(e2.name.startswith('Test Factory Edge'))
        e1.delete()
        e2.delete()
        v1.delete()
        v2.delete()


class BlueprintsWrapperVertex(Vertex):
    element_type = 'blueprints_wrapper_vertex'
    name = properties.String(required=True, max_length=128)


class BlueprintsWrapperEdge(Edge):
    element_type = 'blueprints_wrapper_edge'
    name = properties.String(required=True, max_length=128)


@attr('unit', 'tools', 'blueprintswrapper')
class TestBlueprintsWrapper(BaseMogwaiTestCase):

    def test_blueprints_wrapper(self):
        k = 10
        wrapper_config = {
            'class_name': "ReadOnlyGraph",
            'bindings': {'k': k}
        }
        with BlueprintsWrapper(**wrapper_config):
            gsk = connection.execute_query('"powers of ${k}"')
            pysk = "powers of {}".format(k)
            self.assertEqual(gsk, pysk)
            kk = connection.execute_query("k * k")
            self.assertEqual(kk, k * k)

    @raises(RexProScriptException, MogwaiQueryError)
    def test_wrapper_isolation(self):
        connection.execute_query("k")

    def test_partition_graph(self):
        with PartitionGraph(write='a'):
            v1 = BlueprintsWrapperVertex.create(name="only in a")
            v3 = BlueprintsWrapperVertex.create(name="started in a")

        with PartitionGraph(write='b', read=['a']):
            v2 = BlueprintsWrapperVertex.create(name="only in b")
            e1 = BlueprintsWrapperEdge.create(outV=v2, inV=v1, name="only in b")
            v3.name = "still in a"
            v3.save()

        with PartitionGraph(write='a'):
            v1 = BlueprintsWrapperVertex.get(v1._id)
            self.assertEqual(len(v1.bothE()), 0)
            with self.assertRaises(BlueprintsWrapperVertex.DoesNotExist):
                v2 = BlueprintsWrapperVertex.get(v2._id)

        self.assertEqual(len(v1.bothE()), 1)
        self.assertIsInstance(v1.bothE()[0], BlueprintsWrapperEdge)
        self.assertEqual(v1.bothE()[0]._id, e1._id)
        v3 = BlueprintsWrapperVertex.get(v3._id)
        self.assertEqual(v3['_partition'], 'a')
        e1.delete()
        v1.delete()
        v2.delete()
        v3.delete()


@nottest
def isolation_query(scope):
    wrapper_config = {
        'class_name': "ReadOnlyGraph",
        'bindings': {'scope': scope},
        'pool_size': 5
    }
    scope_values = []
    with BlueprintsWrapper(**wrapper_config) as pool:
        for i in range(7):
            scope_val = connection.execute_query(
                "sleep sleep_length\nreturn scope",
                params={'sleep_length': 100 * (i % 2 + 1) - scope*10},
                pool=pool
            )
            scope_values.append(scope_val)

    return scope, scope_values

@nottest
def nested_wrappers(scope):
    partition_config = {
        'write': "characters",
        'bindings': {'scope': scope},
        'pool_size': 3
    }
    with PartitionGraph(**partition_config) as pool:
        pile = eventlet.GreenPile()
        [pile.spawn(isolation_query, i) for i in range(3)]
        scope_val = connection.execute_query('scope', pool=pool)
        scope_v = BlueprintsWrapperVertex.create(name=scope_val, pool=pool)
        scope_v['nested_values'] = list(pile)
        scope_v.save(pool=pool)

    return scope, scope_v

@attr('unit', 'tools', 'blueprintswrapper', 'concurrency')
class TestWrapperConcurrency(BaseMogwaiTestCase):

    @classmethod
    def setUpClass(cls):
        super(TestWrapperConcurrency, cls).setUpClass()
        host_params = connection.HOST_PARAMS.copy()
        host_params.pop('port')
        connection.setup(concurrency='eventlet', **host_params)

    @classmethod
    def tearDownClass(cls):
        host_params = connection.HOST_PARAMS.copy()
        host_params.pop('port')
        connection.setup(concurrency='sync', **host_params)
        super(TestWrapperConcurrency, cls).tearDownClass()

    def test_wrapper_isolation(self):
        pile = eventlet.GreenPile()
        [pile.spawn(isolation_query, i) for i in range(10)]

        for scope, scope_values in pile:
            for val in scope_values:
                self.assertEqual(scope, val)

    def test_nested_values(self):
        pile = eventlet.GreenPile()
        [pile.spawn(nested_wrappers, chr(i)) for i in range(97, 103)]

        for scope, scope_v in pile:
            self.assertEqual(scope, scope_v.name)
            for inner_scope, scope_values in scope_v['nested_values']:
                for val in scope_values:
                    self.assertEqual(inner_scope, val)
            scope_v.delete()



from __future__ import unicode_literals
from mogwai.exceptions import SaveStrategyException


class Strategy(object):
    """ Saving strategies for mogwai. These are used to indicate when a property
    should be saved after the initial vertex/edge creation.
    """

    @classmethod
    def condition(cls, previous_value, value, has_changed=False, first_save=False, graph_property=None):
        """ Default save strategy condition

        :raises: NotImplementedError
        """
        raise NotImplementedError("No Save Strategy condition defined")

    def __repr__(self):
        return "%s" % self.__class__.__name__

    def __str__(self):
        return self.__repr__()


class SaveOnce(Strategy):
    """ Only save this value once. If it changes throw an exception. """

    @classmethod
    def condition(cls, previous_value, value, has_changed=False, first_save=False, graph_property=None):
        """ Always save this value if it has changed

        :raises: SaveStrategyException
        :rtype: bool
        """
        if not first_save:
            field_name = getattr(graph_property, 'field_db_name', None)
            raise SaveStrategyException("Attempt to change property '%s' with save strategy SAVE_ONCE" % (field_name))
        else:
            return True


class SaveOnChange(Strategy):
    """ Only save this value if it has changed. """

    @classmethod
    def condition(cls, previous_value, value, has_changed=False, first_save=False, graph_property=None):
        """ Always save this value if it has changed

        :rtype: bool
        """
        return has_changed


class SaveAlways(Strategy):
    """ Save this value every time the corresponding model is saved. """

    @classmethod
    def condition(cls, previous_value, value, has_changed=False, first_save=False, graph_property=None):
        """ Save this value every time the corresponding model is saved.

        :rtype: bool
        """
        return True


class SaveOnIncrease(Strategy):
    """ Save this value only if it is increasing """

    @classmethod
    def condition(cls, previous_value, value, has_changed=False, first_save=False, graph_property=None):
        """ Only save this value if it is increasing

        :rtype: bool
        """
        if previous_value is not None:
            return value > previous_value
        else:
            return True


class SaveOnDecrease(Strategy):
    """ Save this value only if it is decreasing """

    @classmethod
    def condition(cls, previous_value, value, has_changed=False, first_save=False, graph_property=None):
        """ Only save this value if it is decreasing

        :rtype: bool
        """
        if previous_value is not None:
            return value < previous_value
        else:
            return True


# common tooling
from pyformance.registry import MetricsRegistry, RegexRegistry
from pyformance.meters import Counter, Histogram, Meter, Timer
from mogwai.exceptions import MogwaiMetricsException

# Default Reporter
from .base import ConsoleMetricReporter



from .vertex import Vertex


def to_offset(page_num, per_page):
    """
    Convert a page_num and per_page to offset.

    :param page_num: The current page number
    :type page_num: int
    :param per_page: The maximum number of results per page
    :type per_page: int
    :rtype: int

    """
    if page_num and per_page:
        return (page_num-1) * per_page
    else:
        return None


class PaginatedVertex(Vertex):
    """
    Convenience class to easily handle pagination for traversals
    """

    @staticmethod
    def _transform_kwargs(kwargs):
        """
        Transforms paginated kwargs into limit/offset kwargs
        """
        values = kwargs.copy()
        return {
            'limit': kwargs.get('per_page'),
            'offset': to_offset(kwargs.get('page_num'), kwargs.get('per_page')),
            'types': kwargs.get('types'),
        }

    __abstract__ = True

    def outV(self, *labels, **kwargs):
        """
        :param labels: pass in the labels to follow in as positional arguments
        :param page_num: the page number to return
        :param per_page: the number of objects to return per page
        :param types: the element types this method is allowed to return
        :rtype: vertex.Vertex
        """
        return super(PaginatedVertex, self).outV(*labels, **self._transform_kwargs(kwargs))

    def outE(self, *labels, **kwargs):
        """
        :param labels: pass in the labels to follow in as positional arguments
        :param page_num: the page number to return
        :param per_page: the number of objects to return per page
        :param types: the element types this method is allowed to return
        :rtype: edge.Edge
        """
        return super(PaginatedVertex, self).outE(*labels, **self._transform_kwargs(kwargs))

    def inV(self, *labels, **kwargs):
        """
        :param labels: pass in the labels to follow in as positional arguments
        :param page_num: the page number to return
        :param per_page: the number of objects to return per page
        :param types: the element types this method is allowed to return
        :rtype: vertex.Vertex
        """
        return super(PaginatedVertex, self).inV(*labels, **self._transform_kwargs(kwargs))

    def inE(self, *labels, **kwargs):
        """
        :param labels: pass in the labels to follow in as positional arguments
        :param page_num: the page number to return
        :param per_page: the number of objects to return per page
        :param types: the element types this method is allowed to return
        :rtype: edge.Edge
        """
        return super(PaginatedVertex, self).inE(*labels, **self._transform_kwargs(kwargs))

    def bothV(self, *labels, **kwargs):
        """
        :param labels: pass in the labels to follow in as positional arguments
        :param page_num: the page number to return
        :param per_page: the number of objects to return per page
        :param types: the element types this method is allowed to return
        :rtype: list[vertex.Vertex]
        """
        return super(PaginatedVertex, self).bothV(*labels, **self._transform_kwargs(kwargs))

    def bothE(self, *labels, **kwargs):
        """
        :param labels: pass in the labels to follow in as positional arguments
        :param page_num: the page number to return
        :param per_page: the number of objects to return per page
        :param types: the element types this method is allowed to return
        :rtype: list[edge.Edge]
        """
        return super(PaginatedVertex, self).bothE(*labels, **self._transform_kwargs(kwargs))